@@ -16,7 +16,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  * Copyright (C) 2008 Novell, Inc.
- * Copyright (C) 2008 Red Hat, Inc.
+ * Copyright (C) 2008 - 2010 Red Hat, Inc.
  */
 
 #include <config.h>
@@ -38,11 +38,10 @@
 #include "nm-system-config-interface.h"
 #include "nm-keyfile-connection.h"
 #include "writer.h"
+#include "common.h"
 
-#define KEYFILE_PLUGIN_NAME "keyfile"
-#define KEYFILE_PLUGIN_INFO "(c) 2007 - 2008 Red Hat, Inc.  To report bugs please use the NetworkManager mailing list."
-
-#define CONF_FILE SYSCONFDIR "/NetworkManager/nm-system-settings.conf"
+#define CONF_FILE SYSCONFDIR "/NetworkManager/NetworkManager.conf"
+#define OLD_CONF_FILE SYSCONFDIR "/NetworkManager/nm-system-settings.conf"
 
 static char *plugin_get_hostname (SCPluginKeyfile *plugin);
 static void system_config_interface_init (NMSystemConfigInterface *system_config_interface_class);
@@ -59,6 +58,7 @@ typedef struct {
 	GFileMonitor *monitor;
 	guint monitor_id;
 
+	const char *conf_file;
 	GFileMonitor *conf_file_monitor;
 	guint conf_file_monitor_id;
 
@@ -72,31 +72,49 @@ read_connections (NMSystemConfigInterface *config)
 {
 	SCPluginKeyfilePrivate *priv = SC_PLUGIN_KEYFILE_GET_PRIVATE (config);
 	GDir *dir;
-	GError *err = NULL;
+	GError *error = NULL;
+	const char *item;
 
-	dir = g_dir_open (KEYFILE_DIR, 0, &err);
-	if (dir) {
-		const char *item;
-
-		while ((item = g_dir_read_name (dir))) {
-			NMKeyfileConnection *connection;
-			char *full_path;
-
-			full_path = g_build_filename (KEYFILE_DIR, item, NULL);
-			connection = nm_keyfile_connection_new (full_path);
-			if (connection) {
-				g_hash_table_insert (priv->hash,
-				                     (gpointer) nm_keyfile_connection_get_filename (connection),
-				                     connection);
-			}
-			g_free (full_path);
-		}
-
-		g_dir_close (dir);
-	} else {
-		g_warning ("Can not read directory '%s': %s", KEYFILE_DIR, err->message);
-		g_error_free (err);
+	dir = g_dir_open (KEYFILE_DIR, 0, &error);
+	if (!dir) {
+		PLUGIN_WARN (KEYFILE_PLUGIN_NAME, "Cannot read directory '%s': (%d) %s",
+		             KEYFILE_DIR,
+		             error ? error->code : -1,
+		             error && error->message ? error->message : "(unknown)");
+		g_clear_error (&error);
+		return;
 	}
+
+	while ((item = g_dir_read_name (dir))) {
+		NMKeyfileConnection *connection;
+		char *full_path;
+
+		full_path = g_build_filename (KEYFILE_DIR, item, NULL);
+		PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "parsing %s ... ", item);
+		connection = nm_keyfile_connection_new (full_path, &error);
+		if (connection) {
+			NMSettingConnection *s_con;
+			const char *cid;
+
+			s_con = (NMSettingConnection *) nm_connection_get_setting (NM_CONNECTION (connection), NM_TYPE_SETTING_CONNECTION);
+			g_assert (s_con);
+
+			cid = nm_setting_connection_get_id (s_con);
+			g_assert (cid);
+
+			g_hash_table_insert (priv->hash,
+			                     (gpointer) nm_keyfile_connection_get_filename (connection),
+			                     connection);
+
+			PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "    read connection '%s'", cid);
+		} else {
+			PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "    error: %s",
+				          (error && error->message) ? error->message : "(unknown)");
+			g_clear_error (&error);
+		}
+		g_free (full_path);
+	}
+	g_dir_close (dir);
 }
 
 typedef struct {
@@ -145,17 +163,33 @@ update_connection_settings (NMKeyfileConnection *orig,
 
 /* Monitoring */
 
+static void
+remove_connection (SCPluginKeyfile *self,
+                   NMKeyfileConnection *connection,
+                   const char *name)
+{
+	g_return_if_fail (connection != NULL);
+	g_return_if_fail (name != NULL);
+
+	/* Removing from the hash table should drop the last reference */
+	g_object_ref (connection);
+	g_hash_table_remove (SC_PLUGIN_KEYFILE_GET_PRIVATE (self)->hash, name);
+	g_signal_emit_by_name (connection, "removed");
+	g_object_unref (connection);
+}
+
 static void
 dir_changed (GFileMonitor *monitor,
-		   GFile *file,
-		   GFile *other_file,
-		   GFileMonitorEvent event_type,
-		   gpointer user_data)
+             GFile *file,
+             GFile *other_file,
+             GFileMonitorEvent event_type,
+             gpointer user_data)
 {
 	NMSystemConfigInterface *config = NM_SYSTEM_CONFIG_INTERFACE (user_data);
 	SCPluginKeyfilePrivate *priv = SC_PLUGIN_KEYFILE_GET_PRIVATE (config);
 	char *name;
 	NMKeyfileConnection *connection;
+	GError *error = NULL;
 
 	name = g_file_get_path (file);
 	connection = g_hash_table_lookup (priv->hash, name);
@@ -163,27 +197,32 @@ dir_changed (GFileMonitor *monitor,
 	switch (event_type) {
 	case G_FILE_MONITOR_EVENT_DELETED:
 		if (connection) {
-			/* Removing from the hash table should drop the last reference */
-			g_object_ref (connection);
-			g_hash_table_remove (priv->hash, name);
-			g_signal_emit_by_name (connection, "removed");
-			g_object_unref (connection);
+			PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "removed %s.", name);
+			remove_connection (SC_PLUGIN_KEYFILE (config), connection, name);
 		}
 		break;
 	case G_FILE_MONITOR_EVENT_CREATED:
 	case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
+		PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "updating %s", name);
+
 		if (connection) {
 			/* Update */
 			NMKeyfileConnection *tmp;
 
-			tmp = (NMKeyfileConnection *) nm_keyfile_connection_new (name);
+			tmp = nm_keyfile_connection_new (name, &error);
 			if (tmp) {
 				update_connection_settings (connection, tmp);
 				g_object_unref (tmp);
+			} else {
+				/* Error; remove the connection */
+				PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "    error: %s",
+						      (error && error->message) ? error->message : "(unknown)");
+				g_clear_error (&error);
+				remove_connection (SC_PLUGIN_KEYFILE (config), connection, name);
 			}
 		} else {
 			/* New */
-			connection = nm_keyfile_connection_new (name);
+			connection = nm_keyfile_connection_new (name, &error);
 			if (connection) {
 				NMSettingConnection *s_con;
 				const char *connection_uuid;
@@ -232,6 +271,10 @@ dir_changed (GFileMonitor *monitor,
 					                     connection);
 					g_signal_emit_by_name (config, NM_SYSTEM_CONFIG_INTERFACE_CONNECTION_ADDED, connection);
 				}
+			} else {
+				PLUGIN_PRINT (KEYFILE_PLUGIN_NAME, "    error: %s",
+						      (error && error->message) ? error->message : "(unknown)");
+				g_clear_error (&error);
 			}
 		}
 		break;
@@ -297,7 +340,7 @@ setup_monitoring (NMSystemConfigInterface *config)
 		priv->monitor = monitor;
 	}
 
-	file = g_file_new_for_path (CONF_FILE);
+	file = g_file_new_for_path (priv->conf_file);
 	monitor = g_file_monitor_file (file, G_FILE_MONITOR_NONE, NULL, NULL);
 	g_object_unref (file);
 
@@ -344,12 +387,13 @@ add_connection (NMSystemConfigInterface *config,
 static GSList *
 get_unmanaged_specs (NMSystemConfigInterface *config)
 {
+	SCPluginKeyfilePrivate *priv = SC_PLUGIN_KEYFILE_GET_PRIVATE (config);
 	GKeyFile *key_file;
 	GSList *specs = NULL;
 	GError *error = NULL;
 
 	key_file = g_key_file_new ();
-	if (g_key_file_load_from_file (key_file, CONF_FILE, G_KEY_FILE_NONE, &error)) {
+	if (g_key_file_load_from_file (key_file, priv->conf_file, G_KEY_FILE_NONE, &error)) {
 		char *str;
 
 		str = g_key_file_get_value (key_file, "keyfile", "unmanaged-devices", NULL);
@@ -366,7 +410,7 @@ get_unmanaged_specs (NMSystemConfigInterface *config)
 			g_free (udis); /* Yes, g_free, not g_strfreev because we need the strings in the list */
 		}
 	} else {
-		g_warning ("Error parsing file '%s': %s", CONF_FILE, error->message);
+		g_warning ("Error parsing file '%s': %s", priv->conf_file, error->message);
 		g_error_free (error);
 	}
 
@@ -378,15 +422,16 @@ get_unmanaged_specs (NMSystemConfigInterface *config)
 static char *
 plugin_get_hostname (SCPluginKeyfile *plugin)
 {
+	SCPluginKeyfilePrivate *priv = SC_PLUGIN_KEYFILE_GET_PRIVATE (plugin);
 	GKeyFile *key_file;
 	char *hostname = NULL;
 	GError *error = NULL;
 
 	key_file = g_key_file_new ();
-	if (g_key_file_load_from_file (key_file, CONF_FILE, G_KEY_FILE_NONE, &error))
+	if (g_key_file_load_from_file (key_file, priv->conf_file, G_KEY_FILE_NONE, &error))
 		hostname = g_key_file_get_value (key_file, "keyfile", "hostname", NULL);
 	else {
-		g_warning ("Error parsing file '%s': %s", CONF_FILE, error->message);
+		g_warning ("Error parsing file '%s': %s", priv->conf_file, error->message);
 		g_error_free (error);
 	}
 
@@ -404,7 +449,7 @@ plugin_set_hostname (SCPluginKeyfile *plugin, const char *hostname)
 	gboolean result = FALSE;
 
 	key_file = g_key_file_new ();
-	if (g_key_file_load_from_file (key_file, CONF_FILE, G_KEY_FILE_NONE, &error)) {
+	if (g_key_file_load_from_file (key_file, priv->conf_file, G_KEY_FILE_NONE, &error)) {
 		char *data;
 		gsize len;
 
@@ -412,7 +457,7 @@ plugin_set_hostname (SCPluginKeyfile *plugin, const char *hostname)
 
 		data = g_key_file_to_data (key_file, &len, &error);
 		if (data) {
-			g_file_set_contents (CONF_FILE, data, len, &error);
+			g_file_set_contents (priv->conf_file, data, len, &error);
 			g_free (data);
 
 			g_free (priv->hostname);
@@ -425,7 +470,7 @@ plugin_set_hostname (SCPluginKeyfile *plugin, const char *hostname)
 			g_error_free (error);
 		}
 	} else {
-		g_warning ("Error parsing file '%s': %s", CONF_FILE, error->message);
+		g_warning ("Error parsing file '%s': %s", priv->conf_file, error->message);
 		g_error_free (error);
 	}
 
@@ -441,6 +486,11 @@ sc_plugin_keyfile_init (SCPluginKeyfile *plugin)
 {
 	SCPluginKeyfilePrivate *priv = SC_PLUGIN_KEYFILE_GET_PRIVATE (plugin);
 
+	if (g_file_test (CONF_FILE, G_FILE_TEST_EXISTS))
+		priv->conf_file = CONF_FILE;
+	else
+		priv->conf_file = OLD_CONF_FILE;
+
 	priv->hostname = plugin_get_hostname (plugin);
 }
 
